import copy

import pytest
from icupy import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 53:
    pytest.skip("ICU4C<53", allow_module_level=True)

from icupy import (
    CurrencyAmount, CurrencyUnit, FieldPosition, FieldPositionIterator,
    Formattable, ICUError, Locale, Measure, MeasureFormat, MeasureUnit,
    ParsePosition, TimeUnit, TimeUnitAmount, UErrorCode, UMeasureFormatWidth,
    UnicodeString,
)


def test_api():
    # [1]
    # icu::MeasureFormat::MeasureFormat(
    #       const Locale &locale,
    #       UMeasureFormatWidth width,
    #       UErrorCode &status
    # )
    fmt1 = MeasureFormat(
        Locale.get_us(),
        UMeasureFormatWidth.UMEASFMT_WIDTH_SHORT)
    fmt2 = MeasureFormat(
        "en-US",
        UMeasureFormatWidth.UMEASFMT_WIDTH_SHORT)
    fmt3 = MeasureFormat(
        Locale.get_us(),
        UMeasureFormatWidth.UMEASFMT_WIDTH_NARROW)

    # [3]
    # icu::MeasureFormat::MeasureFormat(const MeasureFormat &other)
    fmt4 = MeasureFormat(fmt1)

    # UBool icu::Format::operator!=(const Format &other)
    assert not (fmt1 != fmt2)
    assert fmt1 != fmt3
    assert not (fmt1 != fmt4)

    # UBool icu::MeasureFormat::operator==(const Format &other)
    assert fmt1 == fmt2
    assert not (fmt1 == fmt3)
    assert fmt1 == fmt4

    # [1]
    # static MeasureFormat *icu::MeasureFormat::createCurrencyFormat(
    #       const Locale &locale,
    #       UErrorCode &ec
    # )
    fmt4 = MeasureFormat.create_currency_format(Locale.get_us())
    assert isinstance(fmt4, MeasureFormat)

    fmt5 = MeasureFormat.create_currency_format("en-US")
    assert isinstance(fmt5, MeasureFormat)
    assert fmt4 == fmt5

    # [2]
    # static MeasureFormat *icu::MeasureFormat::createCurrencyFormat(
    #       UErrorCode &ec
    # )
    fmt6 = MeasureFormat.create_currency_format()
    assert isinstance(fmt6, MeasureFormat)


def test_clone():
    fmt1 = MeasureFormat("en-US", UMeasureFormatWidth.UMEASFMT_WIDTH_SHORT)

    # MeasureFormat *icu::MeasureFormat::clone()
    fmt2 = fmt1.clone()
    assert isinstance(fmt2, MeasureFormat)
    assert fmt2 == fmt1

    fmt3 = copy.copy(fmt1)
    assert fmt3 == fmt1

    fmt4 = copy.deepcopy(fmt1)
    assert fmt4 == fmt1


def test_currency_amount():
    assert issubclass(CurrencyAmount, Measure)

    # [1]
    # icu::CurrencyAmount::CurrencyAmount(
    #       const Formattable &amount,
    #       ConstChar16Ptr isoCode,
    #       UErrorCode &ec
    # )
    camt1 = CurrencyAmount(Formattable(1), "USD")
    camt2 = CurrencyAmount(Formattable("1.0"), "USD")
    assert camt1 == camt2

    # [2]
    # icu::CurrencyAmount::CurrencyAmount(
    #       double amount,
    #       ConstChar16Ptr isoCode,
    #       UErrorCode &ec
    # )
    camt3 = CurrencyAmount(1, "USD")
    assert camt1 != camt3

    # [3]
    # icu::CurrencyAmount::CurrencyAmount(const CurrencyAmount &other)
    camt4 = CurrencyAmount(camt1)
    assert camt1 == camt4

    # CurrencyAmount *icu::CurrencyAmount::clone()
    camt5 = camt1.clone()
    assert isinstance(camt5, CurrencyAmount)
    assert camt1 == camt5

    # const CurrencyUnit &icu::CurrencyAmount::getCurrency()
    unit = camt1.get_currency()
    assert isinstance(unit, CurrencyUnit)
    assert unit.get_iso_currency() == "USD"

    # const char16_t *icu::CurrencyAmount::getISOCurrency()
    iso_code = camt1.get_iso_currency()
    assert isinstance(iso_code, str)
    assert iso_code == "USD"


def test_format():
    fmt1 = MeasureFormat("en-US", UMeasureFormatWidth.UMEASFMT_WIDTH_WIDE)
    fmt2 = MeasureFormat.create_currency_format("en-US")

    # [2]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt1.format(
        Formattable(Measure(Formattable(3.5), MeasureUnit.create_foot())),
        append_to,
        pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "3.5 feet"

    append_to.remove()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt1.format(
        Formattable(CurrencyAmount(Formattable(1), "USD")),
        append_to,
        pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1.00 US dollars"

    append_to.remove()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt2.format(
        Formattable(CurrencyAmount(Formattable(1), "USD")),
        append_to,
        pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "$1.00"

    append_to.remove()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt1.format(
        Formattable(TimeUnitAmount(Formattable(2), TimeUnit.UTIMEUNIT_DAY)),
        append_to,
        pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "2 days"

    # [3]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    pos_iter = FieldPositionIterator()
    with pytest.raises(ICUError) as exc_info:
        _ = fmt1.format(
            Formattable(Measure(Formattable(3.5), MeasureUnit.create_foot())),
            append_to,
            pos_iter)
    assert exc_info.value.args[0] == UErrorCode.U_UNSUPPORTED_ERROR

    append_to.remove()
    pos_iter = FieldPositionIterator()
    with pytest.raises(ICUError) as exc_info:
        _ = fmt1.format(
            Formattable(CurrencyAmount(Formattable(1), "USD")),
            append_to,
            pos_iter)
    assert exc_info.value.args[0] == UErrorCode.U_UNSUPPORTED_ERROR

    append_to.remove()
    pos_iter = FieldPositionIterator()
    with pytest.raises(ICUError) as exc_info:
        _ = fmt2.format(
            Formattable(CurrencyAmount(Formattable(1), "USD")),
            append_to,
            pos_iter)
    assert exc_info.value.args[0] == UErrorCode.U_UNSUPPORTED_ERROR

    append_to.remove()
    pos_iter = FieldPositionIterator()
    with pytest.raises(ICUError) as exc_info:
        _ = fmt1.format(
            Formattable(TimeUnitAmount(Formattable(2),
                                       TimeUnit.UTIMEUNIT_DAY)),
            append_to,
            pos_iter)
    assert exc_info.value.args[0] == UErrorCode.U_UNSUPPORTED_ERROR

    # [4]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    result = fmt1.format(
        Formattable(Measure(Formattable(3.5), MeasureUnit.create_foot())),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "3.5 feet"

    append_to.remove()
    result = fmt1.format(
        Formattable(CurrencyAmount(Formattable(1), "USD")),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1.00 US dollars"

    append_to.remove()
    result = fmt2.format(
        Formattable(CurrencyAmount(Formattable(1), "USD")),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "$1.00"

    append_to.remove()
    result = fmt1.format(
        Formattable(TimeUnitAmount(Formattable(2), TimeUnit.UTIMEUNIT_DAY)),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "2 days"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 55, reason="ICU4C<55")
def test_format_measure_per_unit():
    fmt = MeasureFormat("en-US", UMeasureFormatWidth.UMEASFMT_WIDTH_SHORT)

    # UnicodeString &icu::MeasureFormat::formatMeasurePerUnit(
    #       const Measure &measure,
    #       const MeasureUnit &perUnit,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format_measure_per_unit(
        Measure(Formattable(50), MeasureUnit.create_pound_force()),
        MeasureUnit.create_square_inch(),
        append_to,
        pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result in ("50 lbf/in\xb2",  # 50 lbf/in²
                      "50 psi")


def test_format_measures():
    fmt = MeasureFormat("fr-FR", UMeasureFormatWidth.UMEASFMT_WIDTH_NARROW)

    # UnicodeString &icu::MeasureFormat::formatMeasures(
    #       const Measure *measures,
    #       int32_t measureCount,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    measures = [
        Measure(Formattable(5), MeasureUnit.create_hour()),
        Measure(Formattable(10), MeasureUnit.create_minute()),
    ]
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format_measures(measures, len(measures), append_to, pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "5h 10min"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 58, reason="ICU4C<58")
def test_get_unit_display_name():
    fmt1 = MeasureFormat("en", UMeasureFormatWidth.UMEASFMT_WIDTH_WIDE)
    fmt2 = MeasureFormat("ja", UMeasureFormatWidth.UMEASFMT_WIDTH_WIDE)

    # UnicodeString icu::MeasureFormat::getUnitDisplayName(
    #       const MeasureUnit &unit,
    #       UErrorCode &status
    # )
    result = fmt1.get_unit_display_name(MeasureUnit.create_year())
    assert isinstance(result, UnicodeString)
    assert result == "years"

    result = fmt2.get_unit_display_name(MeasureUnit.create_year())
    assert isinstance(result, UnicodeString)
    assert result == "\u5E74"  # 年


def test_parse_object():
    fmt1 = MeasureFormat("en-US", UMeasureFormatWidth.UMEASFMT_WIDTH_WIDE)
    fmt2 = MeasureFormat.create_currency_format("en-US")

    # [1]
    # void icu::MeasureFormat::parseObject(
    #       const UnicodeString &source,
    #       Formattable &reslt,
    #       ParsePosition &pos
    # )
    result = Formattable()
    pos = ParsePosition(0)
    fmt1.parse_object(UnicodeString("3.5 feet"), result, pos)
    assert pos.get_index() == 0

    result = Formattable()
    pos = ParsePosition(0)
    fmt1.parse_object("3.5 feet", result, pos)
    assert pos.get_index() == 0

    result = Formattable()
    pos = ParsePosition(0)
    fmt1.parse_object("1.00 US dollars", result, pos)
    assert pos.get_index() == 0

    result = Formattable()
    pos = ParsePosition(0)
    fmt2.parse_object("$1.00", result, pos)
    assert pos.get_index() != 0
    assert pos.get_error_index() == -1
    assert result.get_type() == Formattable.OBJECT
    camt = result.get_object()
    assert isinstance(camt, CurrencyAmount)
    num = camt.get_number()
    assert num.get_type() == Formattable.LONG
    assert num.get_long() == 1
    assert camt.get_iso_currency() == "USD"

    result = Formattable()
    pos = ParsePosition(0)
    fmt1.parse_object("2 days", result, pos)
    assert pos.get_index() == 0

    # [3]
    # void icu::Format::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       UErrorCode &status
    # )
    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt1.parse_object(UnicodeString("3.5 feet"), result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR

    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt1.parse_object("3.5 feet", result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR

    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt1.parse_object("1.00 US dollars", result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR

    result = Formattable()
    fmt2.parse_object("$1.00", result)
    assert result.get_type() == Formattable.OBJECT
    camt = result.get_object()
    assert isinstance(camt, CurrencyAmount)
    num = camt.get_number()
    assert num.get_type() == Formattable.LONG
    assert num.get_long() == 1
    assert camt.get_iso_currency() == "USD"

    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt1.parse_object("2 days", result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR


def test_measure():
    # [1]
    # icu::Measure::Measure(const Formattable &number,
    #                       MeasureUnit *adoptedUnit,
    #                       UErrorCode &ec)
    meas1 = Measure(Formattable(100), MeasureUnit.create_meter())
    meas2 = Measure(Formattable("100.0"), MeasureUnit.create_meter())
    meas3 = Measure(Formattable(0.1), MeasureUnit.create_kilometer())
    assert meas1 == meas2
    assert meas1 != meas3
    assert meas2 != meas3

    # [2]
    # icu::Measure::Measure(const Measure &other)
    meas4 = Measure(meas1)
    assert meas1 == meas4

    # Measure *icu::Measure::clone()
    meas5 = meas4.clone()
    assert isinstance(meas5, Measure)
    assert meas1 == meas5

    # const Formattable &icu::Measure::getNumber()
    num = meas1.get_number()
    assert isinstance(num, Formattable)
    assert num.get_type() == Formattable.LONG
    assert num.get_long() == 100

    # const MeasureUnit &icu::Measure::getUnit()
    unit = meas1.get_unit()
    assert isinstance(unit, MeasureUnit)


def test_time_unit_amount():
    assert issubclass(TimeUnitAmount, Measure)

    # [1]
    # icu::TimeUnitAmount::TimeUnitAmount(
    #       const Formattable &number,
    #       TimeUnit::UTimeUnitFields timeUnitField,
    #       UErrorCode &status
    # )
    tuamt1 = TimeUnitAmount(Formattable(1), TimeUnit.UTIMEUNIT_DAY)

    # [2]
    # icu::TimeUnitAmount::TimeUnitAmount(
    #       double amount,
    #       TimeUnit::UTimeUnitFields timeUnitField,
    #       UErrorCode &status
    # )
    tuamt2 = TimeUnitAmount(1, TimeUnit.UTIMEUNIT_DAY)
    # FIXME: TimeUnitAmount.__eq__(self, other: UObject) is not work.
    # assert tuamt1 == tuamt2

    # icu::TimeUnitAmount::TimeUnitAmount(const TimeUnitAmount &other)
    tuamt3 = TimeUnitAmount(tuamt1)
    assert tuamt1 == tuamt3
    # assert tuamt2 == tuamt3

    # TimeUnitAmount *icu::TimeUnitAmount::clone()
    tuamt4 = tuamt1.clone()
    assert isinstance(tuamt4, TimeUnitAmount)
    assert tuamt1 == tuamt4
    # assert tuamt2 == tuamt4

    # const TimeUnit &icu::TimeUnitAmount::getTimeUnit()
    unit = tuamt1.get_time_unit()
    assert isinstance(unit, TimeUnit)
    assert unit.get_time_unit_field() == TimeUnit.UTIMEUNIT_DAY

    # TimeUnit::UTimeUnitFields icu::TimeUnitAmount::getTimeUnitField()
    assert tuamt1.get_time_unit_field() == TimeUnit.UTIMEUNIT_DAY



import pytest
from icupy import (
    FilteredNormalizer2, Normalizer2, UNormalization2Mode,
    UNormalizationCheckResult, US_INV, U_ICU_VERSION_MAJOR_NUM, UnicodeSet,
    UnicodeString,
)


def test_api():
    # static const Normalizer2 *icu::Normalizer2::getInstance(
    #       const char *packageName,
    #       const char *name,
    #       UNormalization2Mode mode,
    #       UErrorCode &errorCode
    # )
    n2 = Normalizer2.get_instance(
        None,
        "nfc",
        UNormalization2Mode.UNORM2_COMPOSE)

    # UnicodeString &icu::Normalizer2::append(
    #       UnicodeString &first,
    #       const UnicodeString &second,
    #       UErrorCode &errorCode
    # )
    first = UnicodeString("A", -1, US_INV)
    second = UnicodeString("\\u030A", -1, US_INV).unescape()
    third = UnicodeString("B", -1, US_INV)
    result = n2.append(first, second).append(third)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(first)
    assert result == first == "\xC5B"

    first = UnicodeString("A", -1, US_INV)
    result = n2.append(first, "\u030A").append(third)
    assert result == first == "\xC5B"

    # UBool icu::Normalizer2::getDecomposition(
    #       UChar32 c,
    #       UnicodeString &decomposition
    # )
    decomposition = UnicodeString()
    assert n2.get_decomposition(0xc5, decomposition)
    assert decomposition == UnicodeString("A\\u030A", -1, US_INV).unescape()

    # UBool icu::Normalizer2::hasBoundaryAfter(UChar32 c)
    # U+304C: Hiragana Letter Ga
    # U+304B: Hiragana Letter Ka
    # U+3099: Combining Katakana-Hiragana Voiced Sound Mark
    assert n2.has_boundary_after(0x304c)
    assert not n2.has_boundary_after(0x304b)
    assert not n2.has_boundary_after(0x3099)

    # UBool icu::Normalizer2::hasBoundaryBefore(UChar32 c)
    # U+0103: Latin Small Letter A with Breve
    # U+0306: Combining Breve
    assert n2.has_boundary_before(0x0103)
    assert n2.has_boundary_before(ord("a"))
    assert not n2.has_boundary_before(0x0306)

    # UBool icu::Normalizer2::isInert(UChar32 c)
    assert n2.is_inert(0x304c)
    assert not n2.is_inert(0x304b)
    assert not n2.is_inert(0x3099)

    # UBool icu::Normalizer2::isNormalized(
    #       const UnicodeString &s,
    #       UErrorCode &errorCode
    # )
    assert n2.is_normalized(UnicodeString(0x304c))
    assert not n2.is_normalized(
        UnicodeString("\\u304B\\u3099", -1, US_INV).unescape())

    assert n2.is_normalized("\u304c")
    assert not n2.is_normalized("\u304B\u3099")

    # [1]
    # UnicodeString icu::Normalizer2::normalize(
    #       const UnicodeString &src,
    #       UErrorCode &errorCode
    # )
    src = UnicodeString("a\\u0306", -1, US_INV).unescape()
    result = n2.normalize(src)
    assert isinstance(result, UnicodeString)
    assert result == "\u0103"

    result = n2.normalize("a\u0306")
    assert result == "\u0103"

    # [2]
    # UnicodeString &icu::Normalizer2::normalize(
    #       const UnicodeString &src,
    #       UnicodeString &dest,
    #       UErrorCode &errorCode
    # )
    dest = UnicodeString()
    result = n2.normalize(src, dest)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(dest)
    assert result == dest == "\u0103"

    dest.remove()
    result = n2.normalize("a\u0306", dest)
    assert result == dest == "\u0103"

    # UnicodeString &icu::Normalizer2::normalizeSecondAndAppend(
    #       UnicodeString &first,
    #       const UnicodeString &second,
    #       UErrorCode &errorCode
    # )
    first = UnicodeString("A", -1, US_INV)
    second = UnicodeString("a\\u0306", -1, US_INV).unescape()
    third = UnicodeString("b")
    result = n2.normalize_second_and_append(first, second).append(third)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(first)
    assert result == first == "A\u0103b"

    first = UnicodeString("A", -1, US_INV)
    result = n2.normalize_second_and_append(first, "a\u0306").append(third)
    assert result == first == "A\u0103b"

    # UNormalizationCheckResult icu::Normalizer2::quickCheck(
    #       const UnicodeString &s,
    #       UErrorCode &errorCode
    # )
    result = n2.quick_check(UnicodeString(0x304c))
    assert result == UNormalizationCheckResult.UNORM_YES
    result = n2.quick_check(
        UnicodeString("\\u304B\\u3099", -1, US_INV).unescape())
    assert result == UNormalizationCheckResult.UNORM_MAYBE

    result = n2.quick_check("\u304c")
    assert result == UNormalizationCheckResult.UNORM_YES
    result = n2.quick_check("\u304B\u3099")
    assert result == UNormalizationCheckResult.UNORM_MAYBE

    # int32_t icu::Normalizer2::spanQuickCheckYes(
    #       const UnicodeString &s,
    #       UErrorCode &errorCode
    # )
    assert n2.span_quick_check_yes(UnicodeString(0x304c))
    assert not n2.span_quick_check_yes(
        UnicodeString("\\u304B\\u3099", -1, US_INV).unescape())

    assert n2.span_quick_check_yes("\u304c")
    assert not n2.span_quick_check_yes("\u304B\u3099")


def test_filtered_normalizer2():
    # icu::FilteredNormalizer2::FilteredNormalizer2(
    #       const Normalizer2 &n2,
    #       const UnicodeSet &filterSet
    # )
    n2 = Normalizer2.get_instance(
        None,
        "nfc",
        UNormalization2Mode.UNORM2_COMPOSE)
    filter_set = UnicodeSet(UnicodeString("[^\\u00a0-\\u00ff]", -1, US_INV))
    fn2 = FilteredNormalizer2(n2, filter_set)

    decomposition = UnicodeString()
    assert n2.get_decomposition(0xe4, decomposition)
    assert decomposition == UnicodeString("a\\u0308", -1, US_INV).unescape()

    assert n2.get_decomposition(0x100, decomposition)
    assert decomposition == UnicodeString("A\\u0304", -1, US_INV).unescape()

    assert not fn2.get_decomposition(0xe4, decomposition)

    assert fn2.get_decomposition(0x100, decomposition)
    assert decomposition == UnicodeString("A\\u0304", -1, US_INV).unescape()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 49, reason="ICU4C<49")
def test_get_instance():
    # static const Normalizer2 *icu::Normalizer2::getInstance(
    #       const char *packageName,
    #       const char *name,
    #       UNormalization2Mode mode,
    #       UErrorCode &errorCode
    # )
    nfkc1 = Normalizer2.get_instance(
        None,
        "nfkc",
        UNormalization2Mode.UNORM2_COMPOSE)
    assert isinstance(nfkc1, Normalizer2)

    # static const Normalizer2 *icu::Normalizer2::getNFCInstance(
    #       UErrorCode &errorCode
    # )
    nfc = Normalizer2.get_nfc_instance()
    assert isinstance(nfc, Normalizer2)

    # static const Normalizer2 *icu::Normalizer2::getNFDInstance(
    #       UErrorCode &errorCode
    # )
    nfd = Normalizer2.get_nfd_instance()
    assert isinstance(nfd, Normalizer2)

    # static const Normalizer2 *icu::Normalizer2::getNFKCCasefoldInstance(
    #       UErrorCode &errorCode
    # )
    nfkc_cf = Normalizer2.get_nfkc_casefold_instance()
    assert isinstance(nfkc_cf, Normalizer2)

    # static const Normalizer2 *icu::Normalizer2::getNFKCInstance(
    #       UErrorCode &errorCode
    # )
    nfkc2 = Normalizer2.get_nfkc_instance()
    assert isinstance(nfkc2, Normalizer2)

    # static const Normalizer2 *icu::Normalizer2::getNFKDInstance(
    #       UErrorCode &errorCode
    # )
    nfkd = Normalizer2.get_nfkd_instance()
    assert isinstance(nfkd, Normalizer2)

    source = UnicodeString("\\u1E9B\\u0323\\xDF", -1, US_INV).unescape()

    result = nfkc1.normalize(source)
    assert result == UnicodeString("\\u1E69\\xDF", -1, US_INV).unescape()

    result = nfkc2.normalize(source)
    assert result == UnicodeString("\\u1E69\\xDF", -1, US_INV).unescape()

    result = nfkc_cf.normalize(source)
    assert result == UnicodeString("\\u1E69ss", -1, US_INV).unescape()

    result = nfc.normalize(source)
    assert result == UnicodeString(
        "\\u1E9B\\u0323\\xDF",
        -1,
        US_INV).unescape()

    result = nfd.normalize(source)
    assert result == UnicodeString(
        "\\u017F\\u0323\\u0307\\xDF",
        -1,
        US_INV).unescape()

    result = nfkd.normalize(source)
    assert result == UnicodeString(
        "s\\u0323\\u0307\\xDF",
        -1,
        US_INV).unescape()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 49, reason="ICU4C<49")
def test_icu_49():
    n2 = Normalizer2.get_nfc_instance()

    # UChar32 icu::Normalizer2::composePair(
    #       UChar32 a,
    #       UChar32 b
    # )
    assert n2.compose_pair(ord("A"), 0x030a) == 0xc5

    # uint8_t icu::Normalizer2::getCombiningClass(UChar32 c)
    assert n2.get_combining_class(0x3099) == 8

    # UBool icu::Normalizer2::getRawDecomposition(
    #       UChar32 c,
    #       UnicodeString &decomposition
    # )
    decomposition = UnicodeString()
    assert n2.get_raw_decomposition(0xc5, decomposition)
    assert decomposition == UnicodeString("A\\u030A", -1, US_INV).unescape()



from icupy import (
    ConstVoidPtr,
    UBIDI_DEFAULT_LTR, UBIDI_DO_MIRRORING, UBIDI_INSERT_LRM_FOR_NUMERIC,
    UBIDI_LEVEL_OVERRIDE, UBIDI_MAP_NOWHERE,
    UBiDiClassCallbackPtr, UBiDiDirection, UBiDiReorderingMode,
    UBiDiReorderingOption, UCharDirection, UProperty,
    u_get_int_property_max_value, u_strlen,
    ubidi_close, ubidi_count_paragraphs, ubidi_count_runs,
    ubidi_get_base_direction, ubidi_get_class_callback,
    ubidi_get_customized_class, ubidi_get_direction, ubidi_get_length,
    ubidi_get_level_at, ubidi_get_levels,
    ubidi_get_logical_index, ubidi_get_logical_map, ubidi_get_logical_run,
    ubidi_get_paragraph, ubidi_get_paragraph_by_index, ubidi_get_para_level,
    ubidi_get_processed_length, ubidi_get_reordering_mode,
    ubidi_get_reordering_options, ubidi_get_result_length, ubidi_get_text,
    ubidi_get_visual_index, ubidi_get_visual_map, ubidi_get_visual_run,
    ubidi_invert_map, ubidi_is_inverse, ubidi_is_order_paragraphs_ltr,
    ubidi_open, ubidi_open_sized,
    ubidi_order_paragraphs_ltr, ubidi_reorder_logical, ubidi_reorder_visual,
    ubidi_set_class_callback, ubidi_set_context, ubidi_set_inverse,
    ubidi_set_line, ubidi_set_para, ubidi_set_reordering_mode,
    ubidi_set_reordering_options, ubidi_write_reordered, ubidi_write_reverse,
)

from . import gc


def test_api():
    text = "bahrain \u0645\u0635\u0631 kuwait"

    # UBiDi *ubidi_open(void)
    # void ubidi_close(UBiDi *pBiDi)
    with gc(ubidi_open(), ubidi_close) as bidi:
        # const UChar *ubidi_getText(const UBiDi *pBiDi)
        result = ubidi_get_text(bidi)
        assert result is None

        # void ubidi_setPara(UBiDi *pBiDi,
        #                    const UChar *text,
        #                    int32_t length,
        #                    UBiDiLevel paraLevel,
        #                    UBiDiLevel *embeddingLevels,
        #                    UErrorCode *pErrorCode
        # )
        ubidi_set_para(bidi, text, -1, UBIDI_DEFAULT_LTR)

        # int32_t ubidi_countParagraphs(UBiDi *pBiDi)
        assert ubidi_count_paragraphs(bidi) == 1

        # int32_t ubidi_countRuns(UBiDi *pBiDi,
        #                         UErrorCode *pErrorCode
        # )
        assert ubidi_count_runs(bidi) == 3

        # UBiDiDirection ubidi_getBaseDirection(const UChar *text,
        #                                       int32_t length
        # )
        assert ubidi_get_base_direction(text, -1) == UBiDiDirection.UBIDI_LTR

        # UBiDiDirection ubidi_getDirection(const UBiDi *pBiDi)
        assert ubidi_get_direction(bidi) == UBiDiDirection.UBIDI_MIXED

        # int32_t ubidi_getLength(const UBiDi *pBiDi)
        assert ubidi_get_length(bidi) == 18

        # UBiDiLevel ubidi_getLevelAt(const UBiDi *pBiDi,
        #                             int32_t charIndex
        # )
        assert ubidi_get_level_at(bidi, 0) == 0
        # ...
        assert ubidi_get_level_at(bidi, 7) == 0
        assert ubidi_get_level_at(bidi, 8) == 1
        assert ubidi_get_level_at(bidi, 9) == 1
        assert ubidi_get_level_at(bidi, 10) == 1
        assert ubidi_get_level_at(bidi, 11) == 0
        # ...
        assert ubidi_get_level_at(bidi, 17) == 0

        # const UBiDiLevel *ubidi_getLevels(UBiDi *pBiDi,
        #                                   UErrorCode *pErrorCode
        # )
        levels = ubidi_get_levels(bidi)
        assert levels == [
            0, 0, 0, 0, 0, 0, 0, 0,
            1, 1, 1,
            0, 0, 0, 0, 0, 0, 0,
        ]

        # int32_t ubidi_getLogicalIndex(UBiDi *pBiDi,
        #                               int32_t visualIndex,
        #                               UErrorCode *pErrorCode
        # )
        assert ubidi_get_logical_index(bidi, 0) == 0
        # ...
        assert ubidi_get_logical_index(bidi, 7) == 7
        assert ubidi_get_logical_index(bidi, 8) == 10
        assert ubidi_get_logical_index(bidi, 9) == 9
        assert ubidi_get_logical_index(bidi, 10) == 8
        assert ubidi_get_logical_index(bidi, 11) == 11
        # ...
        assert ubidi_get_logical_index(bidi, 17) == 17

        # void ubidi_getLogicalMap(UBiDi *pBiDi,
        #                          int32_t *indexMap,
        #                          UErrorCode *pErrorCode
        # )
        logical_map = ubidi_get_logical_map(bidi)
        assert logical_map == [
            0, 1, 2, 3, 4, 5, 6, 7,
            10, 9, 8,
            11, 12, 13, 14, 15, 16, 17,
        ]

        # void ubidi_getLogicalRun(const UBiDi *pBiDi,
        #                          int32_t logicalPosition,
        #                          int32_t *pLogicalLimit,
        #                          UBiDiLevel *pLevel
        # )
        limit, level = ubidi_get_logical_run(bidi, 0)
        assert limit == 8
        assert level == 0
        limit, level = ubidi_get_logical_run(bidi, 8)
        assert limit == 11
        assert level == 1
        limit, level = ubidi_get_logical_run(bidi, 11)
        assert limit == 18
        assert level == 0

        # int32_t ubidi_getParagraph(const UBiDi *pBiDi,
        #                            int32_t charIndex,
        #                            int32_t *pParaStart,
        #                            int32_t *pParaLimit,
        #                            UBiDiLevel *pParaLevel,
        #                            UErrorCode *pErrorCode
        # )
        result, start, limit, level = ubidi_get_paragraph(bidi, 0)
        assert (result, start, limit, level) == (0, 0, 18, 0)

        # void ubidi_getParagraphByIndex(const UBiDi *pBiDi,
        #                                int32_t paraIndex,
        #                                int32_t *pParaStart,
        #                                int32_t *pParaLimit,
        #                                UBiDiLevel *pParaLevel,
        #                                UErrorCode *pErrorCode
        # )
        start, limit, level = ubidi_get_paragraph_by_index(bidi, 0)
        assert (start, limit, level) == (0, 18, 0)

        # UBiDiLevel ubidi_getParaLevel(const UBiDi *pBiDi)
        assert ubidi_get_para_level(bidi) == 0

        # int32_t ubidi_getProcessedLength(const UBiDi *pBiDi)
        assert ubidi_get_processed_length(bidi) == 18

        # UBiDiReorderingMode ubidi_getReorderingMode(UBiDi *pBiDi)
        assert (ubidi_get_reordering_mode(bidi)
                == UBiDiReorderingMode.UBIDI_REORDER_DEFAULT)

        # uint32_t ubidi_getReorderingOptions(UBiDi *pBiDi)
        assert (ubidi_get_reordering_options(bidi)
                == UBiDiReorderingOption.UBIDI_OPTION_DEFAULT)

        # int32_t ubidi_getResultLength(const UBiDi *pBiDi)
        assert ubidi_get_result_length(bidi) == 18

        # const UChar *ubidi_getText(const UBiDi *pBiDi)
        result = ubidi_get_text(bidi)
        assert result == text

        # int32_t ubidi_getVisualIndex(UBiDi *pBiDi,
        #                              int32_t logicalIndex,
        #                              UErrorCode *pErrorCode
        # )
        assert ubidi_get_visual_index(bidi, 0) == 0
        # ...
        assert ubidi_get_visual_index(bidi, 7) == 7
        assert ubidi_get_visual_index(bidi, 8) == 10
        assert ubidi_get_visual_index(bidi, 9) == 9
        assert ubidi_get_visual_index(bidi, 10) == 8
        assert ubidi_get_visual_index(bidi, 11) == 11
        # ...
        assert ubidi_get_visual_index(bidi, 17) == 17

        # void ubidi_getVisualMap(UBiDi *pBiDi,
        #                         int32_t *indexMap,
        #                         UErrorCode *pErrorCode
        # )
        visual_map = ubidi_get_visual_map(bidi)
        assert visual_map == [
            0, 1, 2, 3, 4, 5, 6, 7,
            10, 9, 8,
            11, 12, 13, 14, 15, 16, 17,
        ]

        # UBiDiDirection ubidi_getVisualRun(UBiDi *pBiDi,
        #                                   int32_t runIndex,
        #                                   int32_t *pLogicalStart,
        #                                   int32_t *pLength
        # )
        result, start, length = ubidi_get_visual_run(bidi, 0)
        assert result == UBiDiDirection.UBIDI_LTR
        assert start == 0
        assert length == 8
        result, start, length = ubidi_get_visual_run(bidi, 1)
        assert result == UBiDiDirection.UBIDI_RTL
        assert start == 8
        assert length == 3
        result, start, length = ubidi_get_visual_run(bidi, 2)
        assert result == UBiDiDirection.UBIDI_LTR
        assert start == 11
        assert length == 7

        # void ubidi_invertMap(const int32_t *srcMap,
        #                      int32_t *destMap,
        #                      int32_t length
        # )
        src_map = [0, 1, 2, 4]
        dest_map = ubidi_invert_map(src_map, len(src_map))
        assert dest_map == [0, 1, 2, UBIDI_MAP_NOWHERE]

        # UBool ubidi_isInverse(UBiDi *pBiDi)
        assert not ubidi_is_inverse(bidi)

        # UBool ubidi_isOrderParagraphsLTR(UBiDi *pBiDi)
        assert not ubidi_is_order_paragraphs_ltr(bidi)

        # void ubidi_orderParagraphsLTR(UBiDi *pBiDi,
        #                               UBool orderParagraphsLTR
        # )
        ubidi_order_paragraphs_ltr(bidi, True)
        ubidi_set_para(bidi, text, -1, UBIDI_DEFAULT_LTR)
        assert ubidi_is_order_paragraphs_ltr(bidi)

        # void ubidi_reorderLogical(const UBiDiLevel *levels,
        #                           int32_t length,
        #                           int32_t *indexMap
        # )
        logical_map2 = ubidi_reorder_logical(levels, len(levels))
        assert logical_map2 == logical_map

        # void ubidi_reorderVisual(const UBiDiLevel *levels,
        #                          int32_t length,
        #                          int32_t *indexMap
        # )
        visual_map2 = ubidi_reorder_visual(levels, len(levels))
        assert visual_map2 == visual_map

        # void ubidi_setInverse(UBiDi *pBiDi,
        #                       UBool isInverse
        # )
        ubidi_set_inverse(bidi, True)
        ubidi_set_para(bidi, text, -1, UBIDI_DEFAULT_LTR)
        assert (ubidi_get_reordering_mode(bidi)
                == UBiDiReorderingMode.UBIDI_REORDER_INVERSE_NUMBERS_AS_L)

        # void ubidi_setReorderingMode(UBiDi *pBiDi,
        #                              UBiDiReorderingMode reorderingMode
        # )
        ubidi_set_reordering_mode(
            bidi,
            UBiDiReorderingMode.UBIDI_REORDER_DEFAULT)
        assert (ubidi_get_reordering_mode(bidi)
                == UBiDiReorderingMode.UBIDI_REORDER_DEFAULT)

        # void ubidi_setReorderingOptions(UBiDi *pBiDi,
        #                                 uint32_t reorderingOptions
        # )
        ubidi_set_reordering_options(
            bidi,
            UBiDiReorderingOption.UBIDI_OPTION_INSERT_MARKS)
        assert (ubidi_get_reordering_options(bidi)
                == UBiDiReorderingOption.UBIDI_OPTION_INSERT_MARKS)


def test_set_class_callback():
    # From icu/source/test/intltest/bidiconf.cpp
    char_from_bidi_class = [
        0x6c,  # U_LEFT_TO_RIGHT
        0x52,  # U_RIGHT_TO_LEFT
        0x33,  # U_EUROPEAN_NUMBER
        0x2d,  # U_EUROPEAN_NUMBER_SEPARATOR
        0x25,  # U_EUROPEAN_NUMBER_TERMINATOR
        0x39,  # U_ARABIC_NUMBER
        0x2c,  # U_COMMON_NUMBER_SEPARATOR
        0x2f,  # U_BLOCK_SEPARATOR
        0x5f,  # U_SEGMENT_SEPARATOR
        0x20,  # U_WHITE_SPACE_NEUTRAL
        0x3d,  # U_OTHER_NEUTRAL
    ]

    # UCharDirection UBiDiClassCallback(const void *context,
    #                                   UChar32 c)
    def _callback1(_context: object, _c: int) -> int:
        assert _context is None
        nonlocal char_from_bidi_class
        if _c in char_from_bidi_class:
            return char_from_bidi_class.index(_c)
        return u_get_int_property_max_value(UProperty.UCHAR_BIDI_CLASS) + 1

    def _callback2(_context: object, _c: int) -> int:
        assert isinstance(_context, list)
        if _c in _context:
            return _context.index(_c)
        return u_get_int_property_max_value(UProperty.UCHAR_BIDI_CLASS) + 1

    with gc(ubidi_open(), ubidi_close) as bidi:
        # void ubidi_getClassCallback(UBiDi *pBiDi,
        #                             UBiDiClassCallback **fn,
        #                             const void **context
        # )
        old_fn1, old_context1 = ubidi_get_class_callback(bidi)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_LEFT_TO_RIGHT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_EUROPEAN_NUMBER

        # void ubidi_setClassCallback(UBiDi *pBiDi,
        #                             UBiDiClassCallback *newFn,
        #                             const void *newContext,
        #                             UBiDiClassCallback **oldFn,
        #                             const void **oldContext,
        #                             UErrorCode *pErrorCode
        # )
        fn2 = UBiDiClassCallbackPtr(_callback1)
        context2 = ConstVoidPtr(None)
        old_fn2, old_context2 = ubidi_set_class_callback(bidi, fn2, context2)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_RIGHT_TO_LEFT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_ARABIC_NUMBER

        old_fn3, old_context3 = ubidi_set_class_callback(bidi,
                                                         old_fn1,
                                                         old_context1)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_LEFT_TO_RIGHT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_EUROPEAN_NUMBER

        fn4 = UBiDiClassCallbackPtr(_callback2)
        context4 = ConstVoidPtr(char_from_bidi_class)
        ubidi_set_class_callback(bidi, fn4, context4)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_RIGHT_TO_LEFT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_ARABIC_NUMBER

        ubidi_set_class_callback(bidi, old_fn2, old_context2)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_LEFT_TO_RIGHT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_EUROPEAN_NUMBER

        ubidi_set_class_callback(bidi, old_fn3, old_context3)
        result = ubidi_get_customized_class(bidi, 0x52)
        assert result == UCharDirection.U_RIGHT_TO_LEFT
        result = ubidi_get_customized_class(bidi, 0x39)
        assert result == UCharDirection.U_ARABIC_NUMBER


def test_set_context():
    with gc(ubidi_open(), ubidi_close) as bidi:
        prologue = "<code>abc \u05d0\u05d1</code>"
        text = "<code>\u05d2\u05d3\u05d4 xyz</code>"
        para_level = 1

        ubidi_set_para(bidi, text, -1, para_level)
        result1 = ubidi_get_logical_map(bidi)

        # void ubidi_setContext(UBiDi *pBiDi,
        #                       const UChar *prologue,
        #                       int32_t proLength,
        #                       const UChar *epilogue,
        #                       int32_t epiLength,
        #                       UErrorCode *pErrorCode
        # )
        ubidi_set_context(bidi, prologue, -1, None, 0)
        ubidi_set_para(bidi, text, -1, para_level)
        result2 = ubidi_get_logical_map(bidi)
        assert result1 != result2


def test_set_line():
    text = "bahrain \u0645\u0635\u0631 kuwait"
    length = u_strlen(text)

    # UBiDi *ubidi_openSized(int32_t maxLength,
    #                        int32_t maxRunCount,
    #                        UErrorCode *pErrorCode
    # )
    # void ubidi_close(UBiDi *pBiDi)
    with gc(ubidi_open_sized(length, 0), ubidi_close) as para, gc(
            ubidi_open_sized(length, 0), ubidi_close) as line:
        ubidi_set_para(para, text, -1, UBIDI_DEFAULT_LTR)
        assert ubidi_get_direction(para) == UBiDiDirection.UBIDI_MIXED

        start = 0
        limit, _ = ubidi_get_logical_run(para, start)
        assert limit == 8

        # void ubidi_setLine(const UBiDi *pParaBiDi,
        #                    int32_t start,
        #                    int32_t limit,
        #                    UBiDi *pLineBiDi,
        #                    UErrorCode *pErrorCode
        # )
        ubidi_set_line(para, start, limit, line)
        assert ubidi_get_direction(line) == UBiDiDirection.UBIDI_LTR

        start = limit
        limit, _ = ubidi_get_logical_run(para, start)
        assert limit == 11

        ubidi_set_line(para, start, limit, line)
        assert ubidi_get_direction(line) == UBiDiDirection.UBIDI_RTL

        start = limit
        limit, _ = ubidi_get_logical_run(para, start)
        assert limit == 18

        ubidi_set_line(para, start, limit, line)
        assert ubidi_get_direction(line) == UBiDiDirection.UBIDI_LTR


def test_set_para():
    text = "car means CAR."
    length = u_strlen(text)

    with gc(ubidi_open_sized(length, 0), ubidi_close) as bidi:
        # void ubidi_setPara(UBiDi *pBiDi,
        #                    const UChar *text,
        #                    int32_t length,
        #                    UBiDiLevel paraLevel,
        #                    UBiDiLevel *embeddingLevels,
        #                    UErrorCode *pErrorCode
        # )
        embedding_levels = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            1 | UBIDI_LEVEL_OVERRIDE,
            1 | UBIDI_LEVEL_OVERRIDE,
            1 | UBIDI_LEVEL_OVERRIDE,
            0,
        ]
        ubidi_set_para(bidi, text, length, UBIDI_DEFAULT_LTR, embedding_levels)

        levels = ubidi_get_levels(bidi)
        assert levels == [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            1, 1, 1,
            0,
        ]

        result = ubidi_get_logical_map(bidi)
        assert result == [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
            12, 11, 10,
            13,
        ]

        result = ubidi_get_visual_map(bidi)
        assert result == [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
            12, 11, 10,
            13,
        ]


def test_write_reordered():
    text = "bahrain \u0645\u0635\u0631 kuwait"
    length = u_strlen(text)

    with gc(ubidi_open_sized(length, 0), ubidi_close) as bidi:
        ubidi_set_para(bidi, text, -1, UBIDI_DEFAULT_LTR)

        # int32_t ubidi_writeReordered(UBiDi *pBiDi,
        #                              UChar *dest,
        #                              int32_t destSize,
        #                              uint16_t options,
        #                              UErrorCode *pErrorCode
        # )
        options = 0
        dest = ubidi_write_reordered(bidi, options)
        assert dest == "bahrain \u0631\u0635\u0645 kuwait"

        # LRM: U+200E
        ubidi_set_inverse(bidi, True)
        ubidi_set_para(bidi, text, -1, UBIDI_DEFAULT_LTR)
        options = UBIDI_INSERT_LRM_FOR_NUMERIC | UBIDI_DO_MIRRORING
        dest = ubidi_write_reordered(bidi, options)
        assert dest == "bahrain \u200e\u0631\u0635\u0645\u200e kuwait"

        # int32_t ubidi_writeReverse(const UChar *src,
        #                            int32_t srcLength,
        #                            UChar *dest,
        #                            int32_t destSize,
        #                            uint16_t options,
        #                            UErrorCode *pErrorCode
        # )
        options = UBIDI_DO_MIRRORING
        dest = ubidi_write_reverse(text, length, options)
        assert dest == "tiawuk \u0631\u0635\u0645 niarhab"



from icupy import (
    AnnualTimeZoneRule, DateTimeRule, InitialTimeZoneRule,
    TimeArrayTimeZoneRule, TimeZoneRule, TimeZoneTransition,
    UCalendarDaysOfWeek, UCalendarMonths, U_MILLIS_PER_HOUR as HOUR,
    UnicodeString,
)


def test_annual_time_zone_rule():
    assert issubclass(AnnualTimeZoneRule, TimeZoneRule)
    assert isinstance(AnnualTimeZoneRule.MAX_YEAR, int)
    assert AnnualTimeZoneRule.MAX_YEAR > 0

    date_time_rule = DateTimeRule(
        UCalendarMonths.UCAL_FEBRUARY,
        11,
        1 * HOUR,
        DateTimeRule.WALL_TIME)

    # [1]
    # icu::AnnualTimeZoneRule::AnnualTimeZoneRule(
    #       const UnicodeString &name,
    #       int32_t rawOffset,
    #       int32_t dstSavings,
    #       const DateTimeRule &dateTimeRule,
    #       int32_t startYear,
    #       int32_t endYear
    # )
    ar1 = AnnualTimeZoneRule(
        UnicodeString("a1"),
        -1 * HOUR,
        1 * HOUR,
        date_time_rule,
        1966,
        AnnualTimeZoneRule.MAX_YEAR)

    # [2]
    # icu::AnnualTimeZoneRule::AnnualTimeZoneRule(
    #       const UnicodeString &name,
    #       int32_t rawOffset,
    #       int32_t dstSavings,
    #       DateTimeRule *dateTimeRule,
    #       int32_t startYear,
    #       int32_t endYear
    # )
    ar2 = AnnualTimeZoneRule(
        UnicodeString("a2"),
        -1 * HOUR,
        1 * HOUR,
        date_time_rule.clone(),
        1966,
        AnnualTimeZoneRule.MAX_YEAR)

    # [3]
    # icu::AnnualTimeZoneRule::AnnualTimeZoneRule(
    #       const AnnualTimeZoneRule &source
    # )
    ar3 = AnnualTimeZoneRule(ar1)

    # AnnualTimeZoneRule *icu::AnnualTimeZoneRule::clone()
    ar4 = ar1.clone()
    assert isinstance(ar4, AnnualTimeZoneRule)

    # UnicodeString &icu::TimeZoneRule::getName(UnicodeString &name)
    name = UnicodeString()
    result = ar1.get_name(name)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "a1"

    # int32_t icu::TimeZoneRule::getRawOffset(void)
    assert ar1.get_raw_offset() == -1 * HOUR

    # int32_t icu::TimeZoneRule::getDSTSavings(void)
    assert ar1.get_dst_savings() == 1 * HOUR

    # int32_t icu::AnnualTimeZoneRule::getEndYear(void)
    assert ar1.get_end_year() == AnnualTimeZoneRule.MAX_YEAR

    # const DateTimeRule *icu::AnnualTimeZoneRule::getRule(void)
    assert ar1.get_rule() == date_time_rule

    # int32_t icu::AnnualTimeZoneRule::getStartYear(void)
    assert ar1.get_start_year() == 1966

    # UBool icu::AnnualTimeZoneRule::getFinalStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, result1 = ar1.get_final_start(-1 * HOUR, 1 * HOUR)
    assert not available
    assert isinstance(result1, (int, float))

    # UBool icu::AnnualTimeZoneRule::getFirstStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, result2 = ar1.get_first_start(-1 * HOUR, 1 * HOUR)
    assert available
    assert isinstance(result2, (int, float))

    # UBool icu::AnnualTimeZoneRule::getNextStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, result3 = ar1.get_next_start(0, -1 * HOUR, 1 * HOUR, False)
    assert available
    assert isinstance(result3, (int, float))

    # UBool icu::AnnualTimeZoneRule::getPreviousStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, result4 = ar1.get_previous_start(0, -1 * HOUR, 1 * HOUR, False)
    assert available
    assert isinstance(result4, (int, float))

    # UBool icu::AnnualTimeZoneRule::getStartInYear(
    #       int32_t year,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, result5 = ar1.get_start_in_year(1966, -1 * HOUR, 1 * HOUR)
    assert available
    assert isinstance(result5, (int, float))

    # UBool icu::AnnualTimeZoneRule::isEquivalentTo(const TimeZoneRule &that)
    assert ar1.is_equivalent_to(ar2)
    assert ar1.is_equivalent_to(ar3)
    assert ar1.is_equivalent_to(ar4)

    # UBool icu::AnnualTimeZoneRule::operator!=(const TimeZoneRule &that)
    assert not (ar1 != ar2)
    assert not (ar1 != ar3)
    assert not (ar1 != ar4)
    assert not (ar2 != ar3)
    assert not (ar2 != ar4)
    assert not (ar3 != ar4)

    # UBool icu::AnnualTimeZoneRule::operator==(const TimeZoneRule &that)
    assert ar1 == ar2 == ar3 == ar4


def test_date_time_rule():
    # [1]
    # icu::DateTimeRule::DateTimeRule(
    #       int32_t month,
    #       int32_t dayOfMonth,
    #       int32_t millisInDay,
    #       TimeRuleType timeType
    # )
    dr1 = DateTimeRule(
        UCalendarMonths.UCAL_FEBRUARY,
        11,
        1 * HOUR,
        DateTimeRule.WALL_TIME)

    # DateRuleType icu::DateTimeRule::getDateRuleType(void)
    assert dr1.get_date_rule_type() == DateTimeRule.DOM

    # int32_t icu::DateTimeRule::getRuleDayOfMonth(void)
    assert dr1.get_rule_day_of_month() == 11

    # int32_t icu::DateTimeRule::getRuleDayOfWeek(void)
    assert dr1.get_rule_day_of_week() == 0

    # int32_t icu::DateTimeRule::getRuleMillisInDay(void)
    assert dr1.get_rule_millis_in_day() == 1 * HOUR

    # int32_t icu::DateTimeRule::getRuleMonth(void)
    assert dr1.get_rule_month() == UCalendarMonths.UCAL_FEBRUARY

    # int32_t icu::DateTimeRule::getRuleWeekInMonth(void)
    assert dr1.get_rule_week_in_month() == 0

    # TimeRuleType icu::DateTimeRule::getTimeRuleType(void)
    assert dr1.get_time_rule_type() == DateTimeRule.WALL_TIME

    # [2]
    # icu::DateTimeRule::DateTimeRule(
    #       int32_t month,
    #       int32_t weekInMonth,
    #       int32_t dayOfWeek,
    #       int32_t millisInDay,
    #       TimeRuleType timeType
    # )
    dr2 = DateTimeRule(
        UCalendarMonths.UCAL_FEBRUARY,
        -1,
        UCalendarDaysOfWeek.UCAL_SATURDAY,
        1 * HOUR,
        DateTimeRule.WALL_TIME)
    assert dr2.get_date_rule_type() == DateTimeRule.DOW
    assert dr2.get_rule_day_of_month() == 0
    assert dr2.get_rule_day_of_week() == UCalendarDaysOfWeek.UCAL_SATURDAY
    assert dr2.get_rule_millis_in_day() == 1 * HOUR
    assert dr2.get_rule_month() == UCalendarMonths.UCAL_FEBRUARY
    assert dr2.get_rule_week_in_month() == -1
    assert dr2.get_time_rule_type() == DateTimeRule.WALL_TIME

    # [3]
    # icu::DateTimeRule::DateTimeRule(
    #       int32_t month,
    #       int32_t dayOfMonth,
    #       int32_t dayOfWeek,
    #       UBool after,
    #       int32_t millisInDay,
    #       TimeRuleType timeType
    # )
    dr3 = DateTimeRule(
        UCalendarMonths.UCAL_FEBRUARY,
        11,
        UCalendarDaysOfWeek.UCAL_SUNDAY,
        True,
        1 * HOUR,
        DateTimeRule.WALL_TIME)
    assert dr3.get_date_rule_type() == DateTimeRule.DOW_GEQ_DOM
    assert dr3.get_rule_day_of_month() == 11
    assert dr3.get_rule_day_of_week() == UCalendarDaysOfWeek.UCAL_SUNDAY
    assert dr3.get_rule_millis_in_day() == 1 * HOUR
    assert dr3.get_rule_month() == UCalendarMonths.UCAL_FEBRUARY
    assert dr3.get_rule_week_in_month() == 0
    assert dr3.get_time_rule_type() == DateTimeRule.WALL_TIME

    # [4]
    # icu::DateTimeRule::DateTimeRule(const DateTimeRule &source)
    dr4 = DateTimeRule(dr1)
    assert dr4.get_date_rule_type() == DateTimeRule.DOM
    assert dr4.get_rule_day_of_month() == 11
    assert dr4.get_rule_day_of_week() == 0
    assert dr4.get_rule_millis_in_day() == 1 * HOUR
    assert dr4.get_rule_month() == UCalendarMonths.UCAL_FEBRUARY
    assert dr4.get_rule_week_in_month() == 0
    assert dr4.get_time_rule_type() == DateTimeRule.WALL_TIME

    # DateTimeRule *icu::DateTimeRule::clone()
    dr5 = dr1.clone()
    assert isinstance(dr5, DateTimeRule)

    # UBool icu::DateTimeRule::operator!=(const DateTimeRule &that)
    assert dr1 != dr2
    assert dr1 != dr3
    assert not (dr1 != dr4)
    assert not (dr1 != dr5)
    assert dr2 != dr3
    assert dr2 != dr4
    assert dr2 != dr5
    assert dr3 != dr4
    assert dr3 != dr5
    assert not (dr4 != dr5)

    # UBool icu::DateTimeRule::operator==(const DateTimeRule &that)
    assert not (dr1 == dr2)
    assert not (dr1 == dr3)
    assert dr1 == dr4
    assert dr1 == dr5
    assert not (dr2 == dr3)
    assert not (dr2 == dr4)
    assert not (dr2 == dr5)
    assert not (dr3 == dr4)
    assert not (dr3 == dr5)
    assert dr4 == dr5


def test_initial_time_zone_rule():
    assert issubclass(InitialTimeZoneRule, TimeZoneRule)

    # [1]
    # icu::InitialTimeZoneRule::InitialTimeZoneRule(
    #       const UnicodeString &name,
    #       int32_t rawOffset,
    #       int32_t dstSavings
    # )
    ir1 = InitialTimeZoneRule(UnicodeString("i1"), -1 * HOUR, 1 * HOUR)

    # [2]
    # icu::InitialTimeZoneRule::InitialTimeZoneRule(
    #       const InitialTimeZoneRule &source
    # )
    ir2 = InitialTimeZoneRule(ir1)

    # InitialTimeZoneRule *icu::InitialTimeZoneRule::clone()
    ir3 = ir1.clone()
    assert isinstance(ir3, InitialTimeZoneRule)

    # UnicodeString &icu::TimeZoneRule::getName(UnicodeString &name)
    name = UnicodeString()
    result = ir1.get_name(name)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "i1"

    # int32_t icu::TimeZoneRule::getRawOffset(void)
    assert ir1.get_raw_offset() == -1 * HOUR

    # int32_t icu::TimeZoneRule::getDSTSavings(void)
    assert ir1.get_dst_savings() == 1 * HOUR

    # UBool icu::InitialTimeZoneRule::getFinalStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, _ = ir1.get_final_start(-1 * HOUR, 1 * HOUR)
    assert not available

    # UBool icu::InitialTimeZoneRule::getFirstStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, _ = ir1.get_first_start(-1 * HOUR, 1 * HOUR)
    assert not available

    # UBool icu::InitialTimeZoneRule::getNextStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, _ = ir1.get_next_start(0, -1 * HOUR, 1 * HOUR, False)
    assert not available

    # UBool icu::InitialTimeZoneRule::getPreviousStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, _ = ir1.get_previous_start(0, -1 * HOUR, 1 * HOUR, False)
    assert not available

    # UBool icu::InitialTimeZoneRule::isEquivalentTo(const TimeZoneRule &that)
    assert ir1.is_equivalent_to(ir2)
    assert ir1.is_equivalent_to(ir3)

    # UBool icu::InitialTimeZoneRule::operator!=(const TimeZoneRule &that)
    assert not (ir1 != ir2)
    assert not (ir1 != ir3)
    assert not (ir2 != ir3)

    # UBool icu::InitialTimeZoneRule::operator==(const TimeZoneRule &that)
    assert ir1 == ir2 == ir3


def test_time_array_time_zone_rule():
    assert issubclass(TimeArrayTimeZoneRule, TimeZoneRule)
    start_times = [0.0, 10000000.0]

    # [1]
    # icu::TimeArrayTimeZoneRule::TimeArrayTimeZoneRule(
    #       const UnicodeString &name,
    #       int32_t rawOffset,
    #       int32_t dstSavings,
    #       const UDate *startTimes,
    #       int32_t numStartTimes,
    #       DateTimeRule::TimeRuleType timeRuleType
    # )
    tr1 = TimeArrayTimeZoneRule(
        UnicodeString("t1"),
        -1 * HOUR,
        1 * HOUR,
        start_times,
        len(start_times),
        DateTimeRule.UTC_TIME)

    # [2]
    # icu::TimeArrayTimeZoneRule::TimeArrayTimeZoneRule(
    #       const TimeArrayTimeZoneRule &source
    # )
    tr2 = TimeArrayTimeZoneRule(tr1)

    # TimeArrayTimeZoneRule *icu::TimeArrayTimeZoneRule::clone()
    tr3 = tr1.clone()
    assert isinstance(tr3, TimeArrayTimeZoneRule)

    # UnicodeString &icu::TimeZoneRule::getName(UnicodeString &name)
    name = UnicodeString()
    result = tr1.get_name(name)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "t1"

    # int32_t icu::TimeZoneRule::getRawOffset(void)
    assert tr1.get_raw_offset() == -1 * HOUR

    # int32_t icu::TimeZoneRule::getDSTSavings(void)
    assert tr1.get_dst_savings() == 1 * HOUR

    # int32_t icu::TimeArrayTimeZoneRule::countStartTimes(void)
    assert tr1.count_start_times() == len(start_times)

    # UBool icu::TimeArrayTimeZoneRule::getStartTimeAt(
    #       int32_t index,
    #       UDate &result
    # )
    valid, result = tr1.get_start_time_at(0)
    assert (valid, result) == (True, start_times[0])

    valid, result = tr1.get_start_time_at(1)
    assert (valid, result) == (True, start_times[1])

    valid, result = tr1.get_start_time_at(2)
    assert not valid

    # DateTimeRule::TimeRuleType icu::TimeArrayTimeZoneRule::getTimeType(void)
    assert tr1.get_time_type() == DateTimeRule.UTC_TIME

    # UBool icu::TimeArrayTimeZoneRule::getFinalStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, result1 = tr1.get_final_start(-1 * HOUR, 1 * HOUR)
    assert available
    assert isinstance(result1, (int, float))

    # UBool icu::TimeArrayTimeZoneRule::getFirstStart(
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UDate &result
    # )
    available, result2 = tr1.get_first_start(-1 * HOUR, 1 * HOUR)
    assert available
    assert isinstance(result2, (int, float))

    # UBool icu::TimeArrayTimeZoneRule::getNextStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, result3 = tr1.get_next_start(0, -1 * HOUR, 1 * HOUR, False)
    assert available
    assert isinstance(result3, (int, float))

    # UBool icu::TimeArrayTimeZoneRule::getPreviousStart(
    #       UDate base,
    #       int32_t prevRawOffset,
    #       int32_t prevDSTSavings,
    #       UBool inclusive,
    #       UDate &result
    # )
    available, result4 = tr1.get_previous_start(0, -1 * HOUR, 1 * HOUR, False)
    assert not available
    assert isinstance(result4, (int, float))

    # UBool icu::TimeArrayTimeZoneRule::isEquivalentTo(
    #       const TimeZoneRule &that
    # )
    assert tr1.is_equivalent_to(tr2)
    assert tr1.is_equivalent_to(tr3)

    # UBool icu::TimeArrayTimeZoneRule::operator!=(const TimeZoneRule &that)
    assert not (tr1 != tr2)
    assert not (tr1 != tr3)
    assert not (tr2 != tr3)

    # UBool icu::TimeArrayTimeZoneRule::operator==(const TimeZoneRule &that)
    assert tr1 == tr2 == tr3


def test_time_zone_transition():
    date_time_rule1 = DateTimeRule(
        UCalendarMonths.UCAL_JULY,
        1,
        UCalendarDaysOfWeek.UCAL_SUNDAY,
        0,
        DateTimeRule.WALL_TIME)
    date_time_rule2 = DateTimeRule(
        UCalendarMonths.UCAL_JANUARY,
        1,
        UCalendarDaysOfWeek.UCAL_SUNDAY,
        0,
        DateTimeRule.WALL_TIME)
    from_ = AnnualTimeZoneRule(
        "from",
        0,
        0,
        date_time_rule1,
        0,
        AnnualTimeZoneRule.MAX_YEAR)
    to = AnnualTimeZoneRule(
        "to",
        0,
        0,
        date_time_rule2,
        0,
        AnnualTimeZoneRule.MAX_YEAR)
    time = 1230681600000.0  # 2008-12-31T00:00:00

    # [1]
    # icu::TimeZoneTransition::TimeZoneTransition(
    #       UDate time,
    #       const TimeZoneRule &from,
    #       const TimeZoneRule &to
    # )
    tzt1 = TimeZoneTransition(time, from_, to)

    # [2]
    # icu::TimeZoneTransition::TimeZoneTransition()
    tzt2 = TimeZoneTransition()

    # [3]
    # icu::TimeZoneTransition::TimeZoneTransition(
    #       const TimeZoneTransition &source
    # )
    tzt3 = TimeZoneTransition(tzt2)

    # TimeZoneTransition *icu::TimeZoneTransition::clone()
    tzt4 = tzt1.clone()
    assert isinstance(tzt4, TimeZoneTransition)

    # const TimeZoneRule *icu::TimeZoneTransition::getFrom(void)
    result = tzt1.get_from()
    assert isinstance(result, AnnualTimeZoneRule)
    assert result == from_

    assert tzt2.get_from() is None

    # UDate icu::TimeZoneTransition::getTime(void)
    assert tzt1.get_time() == time
    assert tzt2.get_time() == 0

    # const TimeZoneRule *icu::TimeZoneTransition::getTo(void)
    result = tzt1.get_to()
    assert isinstance(result, AnnualTimeZoneRule)
    assert result == to

    assert tzt2.get_to() is None

    # UBool icu::TimeZoneTransition::operator!=(const TimeZoneTransition &that)
    assert tzt1 != tzt2
    assert tzt1 != tzt3
    assert not (tzt1 != tzt4)
    assert not (tzt2 != tzt3)
    assert tzt2 != tzt4
    assert tzt3 != tzt4

    # UBool icu::TimeZoneTransition::operator==(const TimeZoneTransition &that)
    assert not (tzt1 == tzt2)
    assert not (tzt1 == tzt3)
    assert tzt1 == tzt4
    assert tzt2 == tzt3
    assert not (tzt2 == tzt4)
    assert not (tzt3 == tzt4)

    # void icu::TimeZoneTransition::setFrom(const TimeZoneRule &from)
    tzt2.set_from(from_)

    # void icu::TimeZoneTransition::setTo(const TimeZoneRule &to)
    tzt2.set_to(to)

    # void icu::TimeZoneTransition::setTime(UDate time)
    tzt2.set_time(time)
    assert tzt1 == tzt2

    tzt3.set_from(from_)
    tzt3.set_to(to)
    tzt3.set_time(time)
    assert tzt1 == tzt3
